import logging
from time import time

from dbtsl.api.shared.query_params import GroupByParam
from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config import Config
from dbt_mcp.prompts.prompts import get_prompt
from dbt_mcp.semantic_layer.client import get_semantic_layer_fetcher
from dbt_mcp.semantic_layer.types import (
    DimensionToolResponse,
    EntityToolResponse,
    MetricToolResponse,
    OrderByParam,
    QueryMetricsSuccess,
)

logger = logging.getLogger(__name__)


def register_sl_tools(dbt_mcp: FastMCP, config: Config) -> None:
    host = config.host
    if not host or not config.token or not config.prod_environment_id:
        raise ValueError(
            "Host, token, and environment ID are required to use semantic layer tools. "
            + "To disable semantic layer tools, "
            + "set DISABLE_SEMANTIC_LAYER=true in your environment."
        )
    semantic_layer_fetcher = get_semantic_layer_fetcher(config)

    @dbt_mcp.tool(description=get_prompt("semantic_layer/list_metrics"))
    def list_metrics() -> list[MetricToolResponse] | str:
        try:
            start_time = time()
            result = semantic_layer_fetcher.list_metrics()
            end_time = time()
            logger.info(f"list_metrics took {end_time - start_time} seconds")
            return result
        except Exception as e:
            logger.error(f"Error listing metrics: {e}")
            return str(e)

    @dbt_mcp.tool(description=get_prompt("semantic_layer/get_dimensions"))
    def get_dimensions(metrics: list[str]) -> list[DimensionToolResponse] | str:
        try:
            start_time = time()
            result = semantic_layer_fetcher.get_dimensions(metrics=metrics)
            end_time = time()
            logger.info(f"get_dimensions took {end_time - start_time} seconds")
            return result
        except Exception as e:
            logger.error(f"Error getting dimensions: {e}")
            return str(e)

    @dbt_mcp.tool(description=get_prompt("semantic_layer/get_entities"))
    def get_entities(metrics: list[str]) -> list[EntityToolResponse] | str:
        try:
            start_time = time()
            result = semantic_layer_fetcher.get_entities(metrics=metrics)
            end_time = time()
            logger.info(f"get_entities took {end_time - start_time} seconds")
            return result
        except Exception as e:
            logger.error(f"Error getting entities: {e}")
            return str(e)

    @dbt_mcp.tool(description=get_prompt("semantic_layer/query_metrics"))
    def query_metrics(
        metrics: list[str],
        group_by: list[GroupByParam] | None = None,
        order_by: list[OrderByParam] | None = None,
        where: str | None = None,
        limit: int | None = None,
    ) -> str:
        try:
            start_time = time()
            result = semantic_layer_fetcher.query_metrics(
                metrics=metrics,
                group_by=group_by,
                order_by=order_by,
                where=where,
                limit=limit,
            )
            end_time = time()
            logger.info(f"query_metrics took {end_time - start_time} seconds")
            if isinstance(result, QueryMetricsSuccess):
                return result.result
            else:
                return result.error
        except Exception as e:
            logger.error(f"Error querying metrics: {e}")
            return str(e)



import subprocess

from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config import Config
from dbt_mcp.prompts.prompts import get_prompt


def register_dbt_cli_tools(dbt_mcp: FastMCP, config: Config) -> None:
    def _run_dbt_command(command: list[str]) -> str:
        process = subprocess.Popen(
            args=[config.dbt_command, *command],
            cwd=config.project_dir,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )
        output, _ = process.communicate()
        return output

    @dbt_mcp.tool(description=get_prompt("dbt_cli/build"))
    def build() -> str:
        return _run_dbt_command(["build"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/compile"))
    def compile() -> str:
        return _run_dbt_command(["compile"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/docs"))
    def docs() -> str:
        return _run_dbt_command(["docs", "generate"])

    @dbt_mcp.tool(name="list", description=get_prompt("dbt_cli/list"))
    def ls() -> str:
        return _run_dbt_command(["list"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/parse"))
    def parse() -> str:
        return _run_dbt_command(["parse"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/run"))
    def run() -> str:
        return _run_dbt_command(["run"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/test"))
    def test() -> str:
        return _run_dbt_command(["test"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/show"))
    def show(sql_query: str, limit: int | None = None) -> str:
        args = ["show", "--inline", sql_query, "--favor-state"]
        if limit:
            args.extend(["--limit", str(limit)])
        args.extend(["--output", "json"])
        return _run_dbt_command(args)



from dataclasses import dataclass

import requests

from dbt_mcp.gql.errors import raise_gql_error


@dataclass
class ConnAttr:
    host: str
    params: dict
    auth_header: str


def submit_request(
    conn_attr: ConnAttr,
    payload: dict,
) -> dict:
    url = f"{conn_attr.host}/api/graphql"
    if "variables" not in payload:
        payload["variables"] = {}
    payload["variables"]["environmentId"] = conn_attr.params["environmentid"]
    r = requests.post(
        url,
        json=payload,
        headers={
            "Authorization": conn_attr.auth_header,
            "x-dbt-partner-source": "dbt-mcp",
        },
    )
    result = r.json()
    raise_gql_error(result)
    return result



from functools import cache

from dbtsl.api.shared.query_params import GroupByParam, OrderByGroupBy
from dbtsl.client.sync import SyncSemanticLayerClient
from dbtsl.error import QueryFailedError

from dbt_mcp.config.config import Config
from dbt_mcp.semantic_layer.gql.gql import GRAPHQL_QUERIES
from dbt_mcp.semantic_layer.gql.gql_request import ConnAttr, submit_request
from dbt_mcp.semantic_layer.levenshtein import get_misspellings
from dbt_mcp.semantic_layer.types import (
    DimensionToolResponse,
    EntityToolResponse,
    MetricToolResponse,
    OrderByParam,
    QueryMetricsError,
    QueryMetricsResult,
    QueryMetricsSuccess,
)


class SemanticLayerFetcher:
    def __init__(self, sl_client: SyncSemanticLayerClient, host: str, config: Config):
        self.sl_client = sl_client
        self.host = host
        self.config = config
        self.entities_cache: dict[str, list[EntityToolResponse]] = {}
        self.dimensions_cache: dict[str, list[DimensionToolResponse]] = {}

    @cache
    def list_metrics(self) -> list[MetricToolResponse]:
        metrics_result = submit_request(
            ConnAttr(
                host=self.host,
                params={"environmentid": self.config.prod_environment_id},
                auth_header=f"Bearer {self.config.token}",
            ),
            {"query": GRAPHQL_QUERIES["metrics"]},
        )
        return [
            MetricToolResponse(
                name=m.get("name"),
                type=m.get("type"),
                label=m.get("label"),
                description=m.get("description"),
            )
            for m in metrics_result["data"]["metrics"]
        ]

    def get_dimensions(self, metrics: list[str]) -> list[DimensionToolResponse]:
        metrics_key = ",".join(sorted(metrics))
        if metrics_key not in self.dimensions_cache:
            dimensions_result = submit_request(
                ConnAttr(
                    host=self.host,
                    params={"environmentid": self.config.prod_environment_id},
                    auth_header=f"Bearer {self.config.token}",
                ),
                {
                    "query": GRAPHQL_QUERIES["dimensions"],
                    "variables": {"metrics": [{"name": m} for m in metrics]},
                },
            )
            dimensions = []
            for d in dimensions_result["data"]["dimensions"]:
                dimensions.append(
                    DimensionToolResponse(
                        name=d.get("name"),
                        type=d.get("type"),
                        description=d.get("description"),
                        label=d.get("label"),
                        granularities=d.get("queryableGranularities")
                        + d.get("queryableTimeGranularities"),
                    )
                )
            self.dimensions_cache[metrics_key] = dimensions
        return self.dimensions_cache[metrics_key]

    def get_entities(self, metrics: list[str]) -> list[EntityToolResponse]:
        metrics_key = ",".join(sorted(metrics))
        if metrics_key not in self.entities_cache:
            entities_result = submit_request(
                ConnAttr(
                    host=self.host,
                    params={"environmentid": self.config.prod_environment_id},
                    auth_header=f"Bearer {self.config.token}",
                ),
                {
                    "query": GRAPHQL_QUERIES["entities"],
                    "variables": {"metrics": [{"name": m} for m in metrics]},
                },
            )
            entities = [
                EntityToolResponse(
                    name=e.get("name"),
                    type=e.get("type"),
                    description=e.get("description"),
                )
                for e in entities_result["data"]["entities"]
            ]
            self.entities_cache[metrics_key] = entities
        return self.entities_cache[metrics_key]

    def validate_query_metrics_params(
        self, metrics: list[str], group_by: list[GroupByParam] | None
    ) -> str | None:
        errors = []
        available_metrics_names = [m.name for m in self.list_metrics()]
        metric_misspellings = get_misspellings(
            targets=metrics,
            words=available_metrics_names,
            top_k=5,
        )
        for metric_misspelling in metric_misspellings:
            recommendations = (
                " Did you mean: " + ", ".join(metric_misspelling.similar_words) + "?"
            )
            errors.append(
                f"Metric {metric_misspelling.word} not found." + recommendations
                if metric_misspelling.similar_words
                else ""
            )

        if errors:
            return f"Errors: {', '.join(errors)}"

        available_dimensions = [d.name for d in self.get_dimensions(metrics)]
        dimension_misspellings = get_misspellings(
            targets=[g.name for g in group_by or []],
            words=available_dimensions,
            top_k=5,
        )
        for dimension_misspelling in dimension_misspellings:
            recommendations = (
                " Did you mean: " + ", ".join(dimension_misspelling.similar_words) + "?"
            )
            errors.append(
                f"Dimension {dimension_misspelling.word} not found." + recommendations
                if dimension_misspelling.similar_words
                else ""
            )

        if errors:
            return f"Errors: {', '.join(errors)}"
        return None

    # TODO: move this to the SDK
    def _format_query_failed_error(self, query_error: Exception) -> QueryMetricsError:
        if isinstance(query_error, QueryFailedError):
            return QueryMetricsError(
                error=str(query_error)
                .replace("QueryFailedError(", "")
                .rstrip(")")
                .lstrip("[")
                .rstrip("]")
                .lstrip('"')
                .rstrip('"')
                .replace("INVALID_ARGUMENT: [FlightSQL]", "")
                .replace("(InvalidArgument; Prepare)", "")
                .replace("(InvalidArgument; ExecuteQuery)", "")
                .replace("Failed to prepare statement:", "")
                .replace(
                    "com.dbt.semanticlayer.exceptions.DataPlatformException:",
                    "",
                )
                .strip()
            )
        else:
            return QueryMetricsError(error=str(query_error))

    def query_metrics(
        self,
        metrics: list[str],
        group_by: list[GroupByParam] | None = None,
        order_by: list[OrderByParam] | None = None,
        where: str | None = None,
        limit: int | None = None,
    ) -> QueryMetricsResult:
        validation_error = self.validate_query_metrics_params(
            metrics=metrics,
            group_by=group_by,
        )
        if validation_error:
            return QueryMetricsError(error=validation_error)

        try:
            query_error = None
            with self.sl_client.session():
                # Catching any exception within the session
                # to ensure it is closed properly
                try:
                    query_result = self.sl_client.query(
                        metrics=metrics,
                        # TODO: remove this type ignore once this PR is merged: https://github.com/dbt-labs/semantic-layer-sdk-python/pull/80
                        group_by=group_by,  # type: ignore
                        order_by=[
                            OrderByGroupBy(
                                name=o.name,
                                descending=o.descending,
                                grain=None,
                            )
                            for o in order_by or []
                        ],
                        where=[where] if where else None,
                        limit=limit,
                    )
                except Exception as e:
                    query_error = e
            if query_error:
                return self._format_query_failed_error(query_error)
            json_result = query_result.to_pandas().to_json(orient="records", indent=2)
            return QueryMetricsSuccess(result=json_result)
        except Exception as e:
            return self._format_query_failed_error(e)


def get_semantic_layer_fetcher(config: Config) -> SemanticLayerFetcher:
    is_local = config.host and config.host.startswith("localhost")
    if is_local:
        host = config.host
    elif config.multicell_account_prefix:
        host = f"{config.multicell_account_prefix}.semantic-layer.{config.host}"
    else:
        host = f"semantic-layer.{config.host}"
    if config.prod_environment_id is None:
        raise ValueError("Environment ID is required")
    if config.token is None:
        raise ValueError("Token is required")
    assert host is not None

    semantic_layer_client = SyncSemanticLayerClient(
        environment_id=config.prod_environment_id,
        auth_token=config.token,
        host=host,
    )

    return SemanticLayerFetcher(
        sl_client=semantic_layer_client,
        host=f"http://{host}" if is_local else f"https://{host}",
        config=config,
    )



import logging

from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config import Config
from dbt_mcp.discovery.client import MetadataAPIClient, ModelsFetcher
from dbt_mcp.prompts.prompts import get_prompt

logger = logging.getLogger(__name__)


def register_discovery_tools(dbt_mcp: FastMCP, config: Config) -> None:
    if not config.host or not config.token or not config.prod_environment_id:
        raise ValueError(
            "Host, token, and environment ID are required to use discovery tools. To disable discovery tools, set DISABLE_DISCOVERY=true in your environment."
        )
    api_client = MetadataAPIClient(
        host=config.host,
        token=config.token,
        multicell_account_prefix=config.multicell_account_prefix,
    )
    models_fetcher = ModelsFetcher(
        api_client=api_client, environment_id=config.prod_environment_id
    )

    @dbt_mcp.tool(description=get_prompt("discovery/get_mart_models"))
    def get_mart_models() -> list[dict] | str:
        try:
            mart_models = models_fetcher.fetch_models(
                model_filter={"modelingLayer": "marts"}
            )
            return [m for m in mart_models if m["name"] != "metricflow_time_spine"]
        except Exception as e:
            logger.error(f"Error getting mart models: {e}")
            return str(e)

    @dbt_mcp.tool(description=get_prompt("discovery/get_all_models"))
    def get_all_models() -> list[dict] | str:
        try:
            return models_fetcher.fetch_models()
        except Exception as e:
            logger.error(f"Error getting all models: {e}")
            return str(e)

    @dbt_mcp.tool(description=get_prompt("discovery/get_model_details"))
    def get_model_details(model_name: str) -> dict | str:
        try:
            return models_fetcher.fetch_model_details(model_name)
        except Exception as e:
            logger.error(f"Error getting model details: {e}")
            return str(e)

    @dbt_mcp.tool(description=get_prompt("discovery/get_model_parents"))
    def get_model_parents(model_name: str) -> list[dict] | str:
        try:
            return models_fetcher.fetch_model_parents(model_name)
        except Exception as e:
            logger.error(f"Error getting model parents: {e}")
            return str(e)



GRAPHQL_QUERIES = {
    "metrics": """
query GetMetrics($environmentId: BigInt!) {
  metrics(environmentId: $environmentId) {
    name
    label
    description
    type
  }
}
    """,
    "dimensions": """
query GetDimensions($environmentId: BigInt!, $metrics: [MetricInput!]!) {
  dimensions(environmentId: $environmentId, metrics: $metrics) {
    description
    name
    type
    queryableGranularities
    queryableTimeGranularities
  }
}
    """,
    "entities": """
query GetEntities($environmentId: BigInt!, $metrics: [MetricInput!]!) {
  entities(environmentId: $environmentId, metrics: $metrics) {
    description
    name
    type
  }
}
    """,
}



import os
from dataclasses import dataclass

from dotenv import load_dotenv


@dataclass
class Config:
    host: str | None
    prod_environment_id: int | None
    dev_environment_id: int | None
    user_id: int | None
    token: str | None
    project_dir: str | None
    dbt_cli_enabled: bool
    semantic_layer_enabled: bool
    discovery_enabled: bool
    remote_enabled: bool
    dbt_command: str
    multicell_account_prefix: str | None
    remote_mcp_url: str


def load_config() -> Config:
    load_dotenv()

    host = os.environ.get("DBT_HOST")
    prod_environment_id = os.environ.get("DBT_PROD_ENV_ID")
    legacy_prod_environment_id = os.environ.get("DBT_ENV_ID")
    dev_environment_id = os.environ.get("DBT_DEV_ENV_ID")
    user_id = os.environ.get("DBT_USER_ID")
    token = os.environ.get("DBT_TOKEN")
    project_dir = os.environ.get("DBT_PROJECT_DIR")
    dbt_path = os.environ.get("DBT_PATH", "dbt")
    disable_dbt_cli = os.environ.get("DISABLE_DBT_CLI", "false") == "true"
    disable_semantic_layer = os.environ.get("DISABLE_SEMANTIC_LAYER", "false") == "true"
    disable_discovery = os.environ.get("DISABLE_DISCOVERY", "false") == "true"
    disable_remote = os.environ.get("DISABLE_REMOTE", "true") == "true"
    multicell_account_prefix = os.environ.get("MULTICELL_ACCOUNT_PREFIX", None)

    errors = []
    if not disable_semantic_layer or not disable_discovery:
        if not host:
            errors.append(
                "DBT_HOST environment variable is required when semantic layer or discovery is enabled."
            )
        if not prod_environment_id and not legacy_prod_environment_id:
            errors.append(
                "DBT_PROD_ENV_ID environment variable is required when semantic layer or discovery is enabled."
            )
        if not token:
            errors.append(
                "DBT_TOKEN environment variable is required when semantic layer or discovery is enabled."
            )
        if host and (host.startswith("metadata") or host.startswith("semantic-layer")):
            errors.append(
                "DBT_HOST must not start with 'metadata' or 'semantic-layer'."
            )
    if not disable_remote:
        if not dev_environment_id:
            errors.append(
                "DBT_DEV_ENV_ID environment variable is required when remote MCP is enabled."
            )
        if not user_id:
            errors.append(
                "DBT_USER_ID environment variable is required when remote MCP is enabled."
            )
    if not disable_dbt_cli:
        if not project_dir:
            errors.append(
                "DBT_PROJECT_DIR environment variable is required when dbt CLI MCP is enabled."
            )
        if not dbt_path:
            errors.append(
                "DBT_PATH environment variable is required when dbt CLI MCP is enabled."
            )

    if errors:
        raise ValueError("Errors found in configuration:\n\n" + "\n".join(errors))

    # Allowing for configuration with legacy DBT_ENV_ID environment variable
    actual_prod_environment_id = (
        int(prod_environment_id)
        if prod_environment_id
        else int(legacy_prod_environment_id)
        if legacy_prod_environment_id
        else None
    )

    return Config(
        host=host,
        prod_environment_id=actual_prod_environment_id,
        dev_environment_id=int(dev_environment_id) if dev_environment_id else None,
        user_id=int(user_id) if user_id else None,
        token=token,
        project_dir=project_dir,
        dbt_cli_enabled=not disable_dbt_cli,
        semantic_layer_enabled=not disable_semantic_layer,
        discovery_enabled=not disable_discovery,
        remote_enabled=not disable_remote,
        dbt_command=dbt_path,
        multicell_account_prefix=multicell_account_prefix,
        remote_mcp_url=(
            "http://" if host and host.startswith("localhost") else "https://"
        )
        + f"{host}/mcp/sse",
    )



def raise_gql_error(result: dict) -> None:
    if result.get("errors"):
        if len(result.get("errors", [])) > 0:
            error_messages = [
                error.get("message", "Unknown error")
                for error in result.get("errors", [])
                if isinstance(error, dict)
            ]
            if error_messages:
                raise ValueError(f"Errors calling API: {', '.join(error_messages)}")
        raise ValueError(
            "Unknown error calling API. "
            + "Check your configuration or contact support if this persists."
        )



import asyncio
import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config import load_config
from dbt_mcp.dbt_cli.tools import register_dbt_cli_tools
from dbt_mcp.discovery.tools import register_discovery_tools
from dbt_mcp.remote.tools import register_remote_tools
from dbt_mcp.semantic_layer.tools import register_sl_tools

config = load_config()
logger = logging.getLogger(__name__)


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[None]:
    logger.info("Starting MCP server")
    try:
        yield
    except Exception as e:
        logger.error(f"Error in MCP server: {e}")
        raise e
    finally:
        logger.info("Shutting down MCP server")


dbt_mcp = FastMCP("dbt", lifespan=app_lifespan)

if config.semantic_layer_enabled:
    register_sl_tools(dbt_mcp, config)

if config.discovery_enabled:
    register_discovery_tools(dbt_mcp, config)

if config.dbt_cli_enabled:
    register_dbt_cli_tools(dbt_mcp, config)

if config.remote_enabled:
    asyncio.run(register_remote_tools(dbt_mcp, config))



import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import (
    Annotated,
    Any,
)

from mcp.client.session import ClientSession
from mcp.client.sse import sse_client
from mcp.server.fastmcp import FastMCP
from mcp.server.fastmcp.tools.base import Tool
from mcp.server.fastmcp.utilities.func_metadata import (
    ArgModelBase,
    FuncMetadata,
    _get_typed_annotation,
)
from mcp.types import EmbeddedResource, ImageContent, TextContent
from mcp.types import Tool as RemoteTool
from pydantic import Field, WithJsonSchema, create_model
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from dbt_mcp.config.config import Config

logger = logging.getLogger(__name__)


@asynccontextmanager
async def sse_mcp_connection_context(
    url: str,
    headers: dict[str, Any] | None = None,
    timeout: float = 5,
    sse_read_timeout: float = 60 * 5,
) -> AsyncGenerator[ClientSession, None]:
    async with (
        sse_client(url, headers, timeout, sse_read_timeout) as (read, write),
        ClientSession(read, write) as session,
    ):
        await session.initialize()
        yield session


# Based on this: https://github.com/modelcontextprotocol/python-sdk/blob/9ae4df85fbab97bf476ddd160b766ca4c208cd13/src/mcp/server/fastmcp/utilities/func_metadata.py#L105
def get_remote_tool_fn_metadata(tool: RemoteTool) -> FuncMetadata:
    dynamic_pydantic_model_params: dict[str, Any] = {}
    for key in tool.inputSchema["properties"].keys():
        # Remote tools shouldn't have type annotations or default values
        # for their arguments. So, we set them to defaults.
        field_info = FieldInfo.from_annotated_attribute(
            annotation=_get_typed_annotation(
                annotation=Annotated[
                    Any,
                    Field(),
                    WithJsonSchema({"title": key, "type": "string"}),
                ],
                globalns={},
            ),
            default=PydanticUndefined,
        )
        dynamic_pydantic_model_params[key] = (field_info.annotation, None)
    return FuncMetadata(
        arg_model=create_model(
            f"{tool.name}Arguments",
            **dynamic_pydantic_model_params,
            __base__=ArgModelBase,
        )
    )


async def list_remote_tools(
    url: str,
    headers: dict[str, Any],
) -> list[RemoteTool]:
    result: list[RemoteTool] = []
    try:
        async with sse_mcp_connection_context(url, headers) as session:
            result = (await session.list_tools()).tools
    except Exception:
        # TODO: uncomment this when remote tools are available
        # and this is actually an error.
        # logger.error(f"Connection error while listing remote tools: {e}")
        pass
    return result


async def register_remote_tools(dbt_mcp: FastMCP, config: Config) -> None:
    assert config.dev_environment_id is not None
    assert config.user_id is not None
    headers = {
        "Authorization": f"Bearer {config.token}",
        "x-dbt-prod-environment-id": str(config.prod_environment_id),
        "x-dbt-dev-environment-id": str(config.dev_environment_id),
        "x-dbt-user-id": str(config.user_id),
    }
    remote_tools = await list_remote_tools(config.remote_mcp_url, headers)
    for tool in remote_tools:
        # Create a new function using a factory to avoid closure issues
        def create_tool_function(tool_name: str):
            async def tool_function(
                *args, **kwargs
            ) -> list[TextContent | ImageContent | EmbeddedResource]:
                async with sse_mcp_connection_context(
                    config.remote_mcp_url, headers
                ) as session:
                    return (
                        await session.call_tool(name=tool_name, arguments=kwargs)
                    ).content

            return tool_function

        new_tool = create_tool_function(tool.name)
        dbt_mcp._tool_manager._tools[tool.name] = Tool(
            fn=new_tool,
            name=tool.name,
            description=tool.description or "",
            parameters=tool.inputSchema,
            fn_metadata=get_remote_tool_fn_metadata(tool),
            is_async=True,
            context_kwarg=None,
        )
